import numpy as np
from tabulate import tabulate
from tqdm import tqdm
from typing import Union


def set_progress_bar(
        max_gen: int, max_nfev: Union[int, None], desc: Union[str, None] = None, unit: Union[str, None] = None) -> tqdm:
    """
    Set the progress bar based on the stopping criterion.

    Args:
        max_gen: Maximum number of generations.
        max_nfev: Maximum number of fitness evaluations.
        desc: Description related to the stopping criterion.
        unit: Unit describing the progress.

    Returns:
        The progress bar.
    """
    if max_nfev is not None:
        pbar = tqdm(total=max_nfev, desc='Fitness Evaluations', unit='nfev', dynamic_ncols=True)
    else:
        desc = desc if desc is not None else 'Generations'
        unit = unit if unit is not None else 'gen'
        pbar = tqdm(total=max_gen, desc=desc, unit=unit, dynamic_ncols=True)
    return pbar


def compute_statistics(fitness: np.ndarray) -> dict:
    """
    Compute the statistics of fitness values.

    Args:
        fitness: Fitness values as an array of real values with (`pop_size`,) shape.

     Returns:
        Dictionary containing the min, max, mean, and standard deviation value.
    """
    stats = {'min': np.min(fitness), 'max': np.max(fitness), 'mean': np.mean(fitness), 'std': np.std(fitness)}
    return stats


def print_info(n_run: int, header: bool = False, **kwargs):
    """
    Print info.

    Args:
        n_run: Independent execution number of the algorithm.
        header: If True, the string indicating the independent execution number is printed.
        kwargs: Information to be printed defined in a dictionary.
    """
    if header:
        print(f'********** Execution #{n_run} **********')
        print(tabulate([kwargs], headers="keys", tablefmt="simple", numalign="left"))
    else:
        formatted_table = [f"{str(item):<{len(k) + 2}}" for k, item in kwargs.items()]
        print("\n" + tabulate([formatted_table], tablefmt="plain"))


class Logbook(dict):
    """
    Class used to store info during the evolution.
    """
    def record(self, **infos):
        """
        Record info.

        Args:
            infos: Info to be stored defined in a dictionary.
        """
        for k, v in infos.items():
            if k in self:
                self[k].append(v)
            else:
                self[k] = [v]

    def get_log(self):
        """
        Get the logbook.
        """
        return self


class BestIndividualTracker:
    """
    Class used to track the best solution ever found during the algorithm execution.
    """
    def __init__(self):
        self.best_ind = None
        self.best_fit = float('inf')

    def update(self, population, fitness):
        """
        Update the tracker.

        Args:
            population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                    shape.
        """
        for ind, fit in zip(population, fitness):
            if fit < self.best_fit:
                self.best_ind = ind
                self.best_fit = fit

    def get_best(self):
        """
        Get the best individual ever found.
        """
        return self.best_ind

    def get_best_fit(self):
        """
        Get the best fitness value ever found.
        """
        return self.best_fit


class FinalResult(dict):
    """
    Class used to store the final result.
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())


class Particle:
    """
    Class used to define a particle described by position, velocity and fitness in :class:`~.PSO` algorithm.

    Args:
        position: A possible solution as array of real parameters with (`n_params`,) shape.
        velocity: Velocity as array of real parameters with (`n_params`,) shape.
        fitness: The corresponding fitness value.
    """
    def __init__(self, position: np.ndarray, velocity: np.ndarray, fitness: float):
        self.position = position
        self.velocity = velocity
        self.fitness = fitness
        self.pbest = position.copy()
        self.fit_pbest = fitness

    def update_velocity(
            self, gbest: np.ndarray, inertia_weight: float, phi1: float, phi2: float, vmin: Union[float, np.ndarray, None],
            vmax: Union[float, np.ndarray, None]):
        """
        Update the velocity of the particle according to the equation proposed in [1].

        References:
            [1] Shi, Y., & Eberhart, R. C., "A modified particle swarm optimizer", Proceedings of the IEEE international
            conference on evolutionary computation, pp. 69–73, 1998.

        Args:
            gbest: Global best position ever visited.
            inertia_weight: Inertia weight.
            phi1: Acceleration coefficient determining the magnitude of the random force in the direction of personal
                  best solution.
            phi2: Acceleration coefficient determining the magnitude of the random force in the direction of global best
                  solution.
            vmin: Lower value(s) of the velocity. If None, no limits are considered.
            vmax: Upper value(s) of the velocity. If None, no limits are considered.
        """
        rand1 = np.random.uniform(0, phi1, size=self.position.size)
        rand2 = np.random.uniform(0, phi2, size=self.position.size)
        new_velocity = inertia_weight * self.velocity + rand1 * (self.pbest - self.position) + rand2 * \
                       (gbest - self.position)

        if vmin is not None and vmax is not None:
            new_velocity[:] = np.clip(new_velocity, vmin, vmax)

        self.velocity = new_velocity

    def update_position(self, param_bounds: Union[tuple, list[tuple]]):
        """
        Update the position of the particle by adding the updated velocity.

        Args:
            param_bounds: Parameter bounds expressed as a tuple (`min`, `max`) or as a list of these tuples of `n_params`
                          size.
        """
        new_position = self.velocity + self.position
        if isinstance(param_bounds, tuple):
            _min, _max = param_bounds
            new_position[:] = np.clip(new_position, _min, _max)
        elif isinstance(param_bounds, list):
            _min = [param_bounds[_][0] for _ in range(len(param_bounds))]
            _max = [param_bounds[_][1] for _ in range(len(param_bounds))]
            new_position[:] = np.clip(new_position, _min, _max)
        self.position = new_position

    def update_pbest(self):
        """
        Update the personal best solution so far.
        """
        if self.fitness < self.fit_pbest:
            self.pbest = self.position.copy()
            self.fit_pbest = self.fitness

    def update_fitness(self, new_fitness: float):
        """
        Update the fitness value.

        Args:
            new_fitness: New fitness value computed.
        """
        self.fitness = new_fitness



import random
import numpy as np
from evovaq.tools.support import print_info, Logbook, FinalResult, set_progress_bar
from evovaq.problem import Problem
from typing import Callable, Union


class HC(object):
    """
    Hill Climbing (HC) is a local search optimization method. In contrast to evolutionary algorithms, HC starts from a
    single initial solution and considers its neighborhood to identify a better solution. Thus, the search takes place
    by iteratively trying to replace the current solution with a better neighbour. There are several variants of Hill
    Climbing search depending on how to find the next solution.
    In `evovaq`, stochastic variant of HC described in [1] is implemented due to its successful application in several
    domains.

    References:
        [1] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "Training circuit-based quantum classifiers through
        memetic algorithms", Pattern Recognition Letters, Elsevier, 2023.

    Args:
        generate_neighbour: A way of generating a neighbour of a current solution. This function is defined as
                            ``generate_neighbour(prob, ind, *args) -> neighbour``, where ``prob`` is :class:`~.Problem`
                            to be solved; ``ind`` is an array of real parameters with (`n_params`,); ``args`` is a tuple
                            of other fixed parameters needed to specify the function; and the output is the neighbour
                            with the same shape as ``ind`` and ``fitness``.
    """
    def __init__(self, generate_neighbour: Callable):
        self.generate_neighbour = generate_neighbour

    def stochastic_var(
            self, problem: Problem, current_solution: np.ndarray, current_fitness: float) -> tuple[np.ndarray, float, int]:
        """
        Stochastic variant.

        Args:
            problem: :class:`~.Problem` to be solved.
            current_solution: A possible solution as array of real parameters with (`n_params`,) shape.
            current_fitness: The corresponding fitness value.

        Returns:
            The improved current solution and fitness value, and the number of fitness evaluations completed during one
            iter.
        """
        neighbour = self.generate_neighbour(problem, current_solution)
        fit_neighbour = problem.evaluate_fitness(neighbour)
        if fit_neighbour < current_fitness:
            current_solution = neighbour.copy()
            current_fitness = fit_neighbour
        return current_solution, current_fitness, 1

    def optimize(
            self, problem: Problem, init_point: Union[None, np.ndarray] = None, max_nfev: Union[int, None] = None,
            max_iter: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            init_point: Initial possible solution as array of real parameters with (`n_params`,) shape. If None, the
                        initial point is randomly generated from `param_bounds`.
            max_nfev: Maximum number of objective function evaluations used as stopping criterion. If None, the maximum
                      number of iterations `max_iter` is considered as stopping criterion.
            max_iter: Maximum number of iterations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the starting solution
        if init_point is None:
            current_solution = problem.generate_individual()
            current_fitness = problem.evaluate_fitness(current_solution)
            nfev = 1
            tot_nfev = 1
        else:
            current_solution = init_point[:].copy()
            current_fitness = problem.evaluate_fitness(current_solution)
            nfev = 1
            tot_nfev = 1

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_iter, max_nfev, 'Iterations', 'iter')
        if max_nfev is not None:
            pbar.update(nfev)

        # Set the logbook
        lg = Logbook()
        lg.record(iter=0, nfev=nfev, fitness=current_fitness)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, iter=0, nfev=nfev, fitness=current_fitness, header=True)

        for it in range(1, max_iter + 1):

            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            current_solution[:], current_fitness, nfev = self.stochastic_var(problem, current_solution, current_fitness)
            tot_nfev += nfev

            # Record info in the logbook
            lg.record(iter=it, nfev=nfev, fitness=current_fitness)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, iter=it, nfev=nfev, fitness=current_fitness)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=current_solution, fun=current_fitness, nfev=tot_nfev, iter=it,
                          log=lg.get_log())
        return res



import matplotlib.pyplot as plt
import numpy as np
import pickle
from openpyxl import Workbook, load_workbook
import pandas as pd
from typing import Union


def show_boxplot(data: Union[list, np.ndarray], xlabels: str, ylabel: str, filename: Union[str, None] = None,
                 colors: Union[str, list, None] = None):
    """
    Show the boxplot of numerical data.

    Args:
        data: Numerical dataset(s).
        xlabels: Labels for each dataset. Length must be compatible with dimensions of `data`.
        ylabel: Label of y-axis.
        filename: Filename of the figure to be saved. If None, the figure is not saved.
        colors: Colors for each boxplot. If None, random colors are generated for each boxplot.
    """
    plt.rcParams['font.size'] = '10'
    plt.figure(figsize=(12, 10))
    plt.ylabel(ylabel)
    basic = dict(linestyle='-', linewidth=1.5, color='black')
    mean = dict(linestyle='--', linewidth=1.5, color='black')
    flier = dict(marker='+', linewidth=1.5, markersize=15)
    bp = plt.boxplot(data, patch_artist=True, showmeans=True, labels=xlabels, meanprops=mean, meanline=True,
                     medianprops=basic, boxprops=basic, whiskerprops=basic, flierprops=flier, capprops=basic,
                     widths=0.3, manage_ticks=True)
    plt.legend([bp['medians'][0], bp['means'][0]], ['Median', 'Mean'], loc='best')

    if colors is not None:
        for patch, color in zip(bp['boxes'], colors):
            patch.set(facecolor=color)
    else:
        rs = np.random.randint(0, 255, len(xlabels))
        gs = np.random.randint(0, 255, len(xlabels))
        bs = np.random.randint(0, 255, len(xlabels))
        colors = [f"#{r:02x}{g:02x}{b:02x}" for r, g, b in zip(rs, gs, bs)]
        for patch, color in zip(bp['boxes'], colors):
            patch.set(facecolor=color)
    plt.show()
    if filename is not None:
        plt.savefig(filename)
    return bp


def write_excel_file(data: dict, filename: str, sheet_title: str, create_new_sheet: bool = False):
    """
    Write an Excel file.
    """
    wb = Workbook()
    if create_new_sheet:
        ws= wb.create_sheet(sheet_title)
    else:
        ws = wb.active
        ws.title = sheet_title
    ws.append(list(data.keys()))
    for row in zip(*data.values()):
        ws.append(row)
    wb.save(filename)
    wb.close()


def read_excel_file(filename: str, sheet_title: Union[str, None] = None):
    """
    Read an Excel file.
    """
    wb = load_workbook(filename=filename)
    if sheet_title is not None:
        ws = wb[sheet_title]
    else:
        ws = wb.active
    data = {}
    for n, col in enumerate(ws.iter_cols(values_only=True)):
        data[f'Column #{n}'] = col
    return data


def write_csv_file(data: dict, filename: str):
    """
    Write a csv file.
    """
    df = pd.DataFrame(data)
    df.to_csv(filename, index=True)


def read_csv_file(filename: str):
    """
    Read a csv file.
    """
    data = pd.read_csv(filename)
    return data


def write_pkl_file(data: dict, filename: str):
    """
    Write a pickle file.
    """
    with open(filename, 'wb') as f:
        pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)


def read_pkl_file(filename: str):
    """
    Read a pickle file.
    """
    with open(filename, 'rb') as f:
        data = pickle.load(f)
    return data



import random
import numpy as np
from functools import partial
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult
from evovaq.tools.operators import sel_best, sel_random
from typing import Callable, Union
from evovaq.problem import Problem


class MA(object):
    """
    Memetic Algorithm (MA) [1] is an evolutionary approach that merge population- and local-based methods to improve
    exploration and exploitation capabilities in visiting the problem search space.
    In `evovaq`, MA workflow described in [2] is implemented.

    References:
        [1] P. Moscato, et al., "On evolution, search, optimization, genetic algorithms and martial arts: towards memetic
        algorithms", Caltech concurrent computation program, C3P Report, vol. 826, pp. 37, 1989.

        [2] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "Training circuit-based quantum classifiers through
        memetic algorithms", Pattern Recognition Letters, Elsevier, 2023.

    Args:
        global_search: Population-based method used to evolve the population of individuals. The global search function
                       is defined as ``global_search(prob, pop, fitness, *args) -> (offspring, fit_offspring, nfev)``, where
                       ``prob`` is :class:`~.Problem` to be solved; ``pop`` and ``fitness`` are two arrays of real parameters
                       with (`pop_size`,`n_params`) and (`n_params`,) shape, respectively; ``args`` is a tuple of other
                       fixed parameters needed to specify the function; and the output is the resulting offspring and
                       fitness values  with the same shape as ``pop`` and ``fitness``, and number of fitness evaluations
                       completed during the evolution. Here, it is possible to use :class:`~.GA` or :class:`~.DE`
                       as a population-based method via :meth:`~DE.evolve_population` method.
        local_search: Local search method used to improve a subset of individuals. The local search function is defined
                      as ``local_search(prob, ind, fitness, *args) -> (ind, fitness, nfev)``, where ``prob`` is
                      :class:`~.Problem` to be solved; ``ind``  is an array of real parameters with (`n_params`,);
                      ``fitness`` is a float value; ``args`` is a tuple of other fixed parameters needed to specify the
                      function; and the output is the improved individual and fitness value with the same shape as
                      ``ind`` and ``fitness``, and number of fitness evaluations completed during the local research.
                      Here, it is possible to use :class:`~.HC` as a local search method via :meth:`~HC.stochastic_var`
                      method.
        sel_for_refinement: Selection operator used to choose the individuals undergoing local refinement.
        frequency: Individual learning frequency defined in the range (0 , 1) influencing the number of individuals that
                   is undergone to local refinement.
        intensity: Individual learning intensity representing the maximum computational budget allowable for individual
                   learning to expend on improving a single solution. Here, it corresponds to the maximum number
                   of iterations to be performed during the local search.
        elitism: If True, the best solution of current population is transferred directly into the next generation.
    """
    def __init__(
            self, global_search: Callable, local_search: Callable, sel_for_refinement: Callable, frequency: float,
            intensity: int, elitism: bool = True):
        self.global_search = global_search
        self.local_search = local_search
        self.sel_for_refinement = sel_for_refinement
        self.frequency = frequency
        self.intensity = intensity
        self.elitism = elitism

        if self.sel_for_refinement == sel_random:
            self.sel_for_refinement = partial(sel_random, replace=False)

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            best, best_fit = sel_best(population, fitness, 1)

            # Global search method used to evolve the population of possible solutions
            offspring, fit_offspring, glob_nfev = self.global_search(problem, population, fitness)
            tot_nfev += glob_nfev
            nfev = glob_nfev

            # Local search method used to improve a subset of individuals
            omega_idx, fit_omega = self.sel_for_refinement(np.arange(len(offspring)), fit_offspring,
                                                           int(self.frequency * pop_size))

            for ind_idx in omega_idx:
                for _ in range(self.intensity):
                    offspring[ind_idx][:], fit_offspring[ind_idx], loc_nfev = self.local_search(problem,
                                                                                                offspring[ind_idx],
                                                                                                fit_offspring[ind_idx])
                    tot_nfev += loc_nfev
                    nfev += loc_nfev

            if self.elitism:
                joint_pop = np.concatenate((offspring, best))
                joint_fitness = np.concatenate((fit_offspring, best_fit))

                # Sort from the best (min fitness) to worst (max fitness) individual
                sorted_idx = np.argsort(joint_fitness)

                # Replacement
                population[:] = joint_pop[sorted_idx[:pop_size]]
                fitness[:] = joint_fitness[sorted_idx[:pop_size]]
            else:
                population[:] = offspring
                fitness[:] = fit_offspring

            # Store the best solution ever found
            best_tracker.update(population, fitness)

            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res



import numpy as np
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult
from evovaq.tools.operators import sel_best
import random
from typing import Union
from evovaq.problem import Problem


class BBBC(object):
    """
    The Big Bang - Big Crunch (BBBC) method as developed by Erol and Eks in 2006 [1] consists of two alternating steps:
    1) a Big Bang phase, where the candidate solutions are randomly distributed over the search space; 2) a Big Crunch
    phase, where a contraction operation estimates a weighted average, denoted as Centre of Mass, of the randomly
    distributed candidate solutions. During the Big Bang phases, new candidate solutions are generated considering the
    Center of Mass and the best global solution, as introduced in [2].

    References:
        [1] O. K. Erol and I. Eksin, “A new optimization method: big bang–big crunch”, Advances in Engineering Software,
        vol. 37, no. 2, pp. 106– 111, 2006.

        [2] C. V. Camp, “Design of space trusses using big bang–big crunch optimization,” Journal of Structural
        Engineering, vol. 133, no. 7, pp. 999–1008, 2007.

    Args:
        elitism: If True, the best solution of current population is transferred directly into the next generation.
        alpha: Hyperparameter limiting the size of the search space.
        beta: Hyperparameter defined in range (0,1) controlling the influence of the best individual on the location of
              new candidate solutions.
    """

    def __init__(self, elitism: bool = True, alpha: float = 10.0, beta: float = 0.25):
        self.elitism = elitism
        self.alpha = alpha
        self.beta = beta

    @staticmethod
    def compute_centre_of_mass(population: np.ndarray, fitness: np.ndarray) -> np.ndarray:
        """
        Compute the centre of mass.

        Args:
            population: A population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                     shape.

        Returns:
            The centre of mass as an array of real parameters with (`n_params`,) shape.
        """
        return np.sum(population / fitness[:, None], axis=0) / np.sum(1 / fitness)

    def generate_new_individual(
            self, problem: Problem, com: np.ndarray, best: np.ndarray, gen: int) -> np.ndarray:
        """
        Generate a new individual distributed between the center of mass and the best global solution.

        Args:
            problem: :class:`~.Problem` to be solved.
            com: The centre of mass as an array of real parameters with (`n_params`,) shape.
            best: The global best solution as an array of real parameters with (`n_params`,) shape.
            gen: Generation number.

        Returns:
            A new individual as an array of real parameters with (`n_params`,) shape.
        """
        r = np.random.normal(0, 1)
        diff = np.zeros(com.shape)
        if isinstance(problem.param_bounds, tuple):
            diff[:] = np.full(len(com), problem.param_bounds[1] - problem.param_bounds[0])
        elif isinstance(problem.param_bounds, list):
            diff[:] = np.array([b[1] - b[0] for b in problem.param_bounds])
        return self.beta * com + (1 - self.beta) * best + diff * r * self.alpha / gen

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Big Bang phase: create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            # Select the best individual of the current population
            best, best_fit = sel_best(population, fitness, 1)

            # Big Crunch phase: compute the Centre of Mass
            com = self.compute_centre_of_mass(population, fitness)

            # Clone the individuals to produce the offspring
            offspring = population.copy()

            # Big Bang phase: create new solutions around the Centre of Mass
            for child in offspring:
                child[:] = self.generate_new_individual(problem, com, best, gen)

            # Check the bounds of the parameters
            for child in offspring:
                child[:] = problem.check_bounds(child)

            # Evaluate the individuals with an invalid fitness
            fit_offspring = np.array(list(map(problem.evaluate_fitness, offspring)))
            nfev = len(offspring)
            tot_nfev += len(offspring)

            if self.elitism:
                joint_pop = np.concatenate((offspring, best))
                joint_fitness = np.concatenate((fit_offspring, best_fit))

                # Sort from the best (min fitness) to worst (max fitness) individual
                sorted_idx = np.argsort(joint_fitness)

                # Replacement
                population[:] = joint_pop[sorted_idx[:pop_size]]
                fitness[:] = joint_fitness[sorted_idx[:pop_size]]
            else:
                population[:] = offspring
                fitness[:] = fit_offspring

            # Store the best solution ever found
            best_tracker.update(population, fitness)

            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res



import numpy as np
from typing import Union, Callable


class Problem:
    """
    Class used to define the minimization problem to be solved.

    Args:
        n_params: Number of real parameters to be optimized.
        param_bounds: Parameter bounds expressed as a tuple (`min`, `max`) or as a list of these tuples of `n_params`
                      size.
        obj_function: The objective function to be minimized defined as ``obj_function(params, *args) -> float`` ,
                       where ``params`` is an array with (`n_params`,) shape and ``args`` is a tuple of other fixed
                       parameters needed to specify the function.
    """

    def __init__(self, n_params: int, param_bounds: Union[tuple, list[tuple]], obj_function: Callable):
        self.n_params = n_params
        self.param_bounds = param_bounds
        self.obj_function = obj_function

    def generate_individual(self) -> np.ndarray:
        """
        Generate a random possible solution, called individual.

        Returns:
            A possible solution randomly generated from `param_bounds`.
        """
        if isinstance(self.param_bounds, tuple):
            _min, _max = self.param_bounds
            individual = np.random.uniform(low=_min, high=_max, size=self.n_params)
        elif isinstance(self.param_bounds, list):
            _min = [self.param_bounds[_][0] for _ in range(len(self.param_bounds))]
            _max = [self.param_bounds[_][1] for _ in range(len(self.param_bounds))]
            individual = np.random.uniform(low=_min, high=_max, size=self.n_params)
            if not len(individual) == self.n_params:
                raise ValueError(
                    "Please insert the bounds as a tuple (min, max) or a list of tuples of length n_params")
        else:
            raise ValueError("Please insert the bounds as a tuple (min, max) or a list of tuples of length n_params")
        return individual

    def generate_random_pop(self, pop_size: int) -> np.ndarray:
        """
        Generate a population of possible random solutions.

        Args:
            pop_size: Population size.

        Returns:
            A set of possible solutions randomly generated from `param_bounds`.
        """
        if isinstance(self.param_bounds, tuple):
            _min, _max = self.param_bounds
            population = np.random.uniform(low=_min, high=_max, size=(pop_size, self.n_params))
        elif isinstance(self.param_bounds, list):
            _min = [self.param_bounds[_][0] for _ in range(len(self.param_bounds))]
            _max = [self.param_bounds[_][1] for _ in range(len(self.param_bounds))]
            population = np.random.uniform(low=_min, high=_max, size=(pop_size, self.n_params))
            if not len(population[0]) == self.n_params:
                raise ValueError(
                    "Please insert the bounds as a tuple (min, max) or a list of tuples of length n_params")
        else:
            raise ValueError(
                "Please insert the bounds as a tuple (min, max) or a list of tuples of length n_params")
        return population

    def evaluate_fitness(self, params: np.ndarray) -> float:
        """
        Evaluate the fitness function of the given parameters.

        Args:
            params: A possible solution as array of real parameters with (`n_params`,) shape.

        Returns:
            Value of the objective function.
        """
        return self.obj_function(params)

    def check_bounds(self, params: np.ndarray) -> np.ndarray:
        """
        Check if the solution `params` satisfies the parameters bounds set in `param_bounds`.

        Args:
            params: A possible solution as array of real parameters with (`n_params`,) shape.

        Returns:
            A possible solution with clipped values according to `param_bounds`.
        """
        if isinstance(self.param_bounds, tuple):
            _min, _max = self.param_bounds
            params[:] = np.clip(params, _min, _max)
        elif isinstance(self.param_bounds, list):
            _min = [self.param_bounds[_][0] for _ in range(len(self.param_bounds))]
            _max = [self.param_bounds[_][1] for _ in range(len(self.param_bounds))]
            params[:] = np.clip(params, _min, _max)
        return params



import numpy as np
from functools import partial
import random
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult
from evovaq.tools.operators import sel_permutation, sel_best, CROSSOVER_ARGS
from evovaq.problem import Problem
from typing import Union, Callable


class CHC(object):
    """
    Cross generational elitist selection, Heterogeneous recombination, and Cataclysmic mutation (CHC) algorithm is a
    nontraditional genetic algorithm which combines a conservative selection strategy that always
    preserves the best individuals found so far with a radical (highly disruptive) crossover operator that produces
    offspring that are maximally different from both parents [1]. In detail, it is based on four main components:
    a elitist selection, a highly disruptive crossover, an incest prevention check to avoid the recombination of similar
    solutions, and a population reinitialization method when the population has converged.
    In `evovaq`, a real-coded CHC version based on [2-3] is implemented. However, the similarity is measured between
    fitness values and not between individuals, since different solutions can lead to equal or close cost values when
    training variational quantum circuits.

    References:
        [1] Larry J. Eshelman, "The CHC Adaptive Search Algorithm: How to Have Safe Search When Engaging in
        Nontraditional Genetic Recombination", Foundations of Genetic Algorithms, Elsevier, vol. 1, pp. 265-283, 1991.

        [2] O. Cordón, S. Damas, J. Santamaría, "Feature-based image registration by means of the CHC evolutionary algorithm",
        Image and Vision Computing, vol. 24, Issue 5, pp. 525-533, 2006.

        [3] Cuéllar, M. P., Gómez-Torrecillas, J., Lobillo, F. J., & Navarro, G., "Genetic algorithms with
        permutation-based representation for computing the distance of linear codes", Swarm and Evolutionary Computation,
        vol. 60, pp. 100797, 2021.

    Args:
        crossover: Crossover operator used to mate two individuals. The crossover function is defined as
                   ``cx_function(par1, par2, *args) -> (child1, child2)``, where ``par1`` and ``par2`` are two arrays of
                   real parameters with (`n_params`,) shape, ``args`` is a tuple of other fixed parameters needed to
                   specify the function, and the output is the resulting children with the same shape as the parents.
        distance: Distance metric used to compute the similarity between parents. In this implementation, similarity is
                  measured between two fitness values, so Euclidean distance can be considered.
        multiplier: Factor influencing the initial crossover threshold.
        dec_percentage:  Crossover threshold update rate.
        kwargs: Additional keyword arguments used to set hyperparameter values of crossover operator.
    """
    def __init__(self,
                 crossover: Callable, distance: Callable, multiplier: float = 1, dec_percentage: float = 0.1, **kwargs):
        self.distance = distance
        self.multiplier = multiplier
        self.dec_percentage = dec_percentage
        self.kwargs = kwargs

        if crossover in CROSSOVER_ARGS:
            args = {arg: self.kwargs.get(arg) for arg in CROSSOVER_ARGS[crossover] if arg in self.kwargs.keys()}
            self.crossover = partial(crossover, **args)
        else:
            self.crossover = crossover

    def incest_prevention_check(self, parents: np.ndarray, fitness: np.ndarray, thr: float) -> tuple[np.ndarray, np.ndarray]:
        """
        Incest prevention check. Before mating, the similarity between the potential parents is calculated, and if this
        distance does not exceed the threshold `thr`, they are not mated.

        Args:
            parents: A set of parents as array of real parameters with (`pop_size`, `n_params`) shape.
            fitness: A set of fitness values associated to the parents as array of real values with (`pop_size`,)
                    shape.
            thr: Crossover threshold.

        Returns:
            Parents and corresponding fitness values allowed for mating.
        """
        allowed = np.zeros(len(parents), dtype=bool)
        for i in range(0, len(parents), 2):
            if self.distance(fitness[i], fitness[i + 1]) > thr:
                allowed[i] = allowed[i + 1] = True
            else:
                pass
        return parents[allowed], fitness[allowed]

    def initialize_cx_threshold(self, population: np.ndarray, fitness: np.ndarray) -> tuple[float, float]:
        """
        Initialize the crossover threshold and compute the decrement value.

        Args:
            population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                    shape.

        Returns:
            The initial crossover threshold and the decrement value.
        """
        avg_d = 0
        n = 0
        max_d = 0
        for i in range(len(population) - 1):
            for j in range(i + 1, len(population)):
                n += 1
                d = self.distance(fitness[i], fitness[j])
                avg_d += d
                if d > max_d:
                    max_d = d
        avg_d /= n
        return avg_d * self.multiplier, max_d * self.dec_percentage

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        # Initialize threshold
        thr, dec = self.initialize_cx_threshold(population, fitness)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            # Parent selection
            parents, fit_parents = sel_permutation(population, fitness)

            # Incest prevention
            allowed_parents, fit_allowed_parents = self.incest_prevention_check(parents, fit_parents, thr)

            # Crossover and replacement
            if len(allowed_parents) > 0:
                # Clone the individuals to produce the offspring
                offspring = allowed_parents.copy()

                for child1, child2 in zip(offspring[::2], offspring[1::2]):
                    child1[:], child2[:] = self.crossover(child1, child2)

                # Check the bounds of the parameters
                for child in offspring:
                    child[:] = problem.check_bounds(child)

                # Evaluate the individuals of the offspring
                fit_offspring = np.array(list(map(problem.evaluate_fitness, offspring)))
                nfev = len(offspring)
                tot_nfev += len(offspring)

                # Elitism
                joint_pop = np.concatenate((population, offspring))
                joint_fitness = np.concatenate((fitness, fit_offspring))

                # Sort from the best (min fitness) to worst (max fitness) individual
                sorted_idx = np.argsort(joint_fitness)

                # Replacement
                population[:] = joint_pop[sorted_idx[:pop_size]]
                fitness[:] = joint_fitness[sorted_idx[:pop_size]]

            else:
                sorted_idx = None
                nfev = 0

            # Decrease the distance threshold
            if sorted_idx is None or np.all(sorted_idx[:pop_size] < pop_size):
                thr -= dec
            # Reinitialization process: if thr <= 0, all parents, even equal, combine and thus there is no exploration
                if thr <= 0:
                    best, best_fit = sel_best(population, fitness, 1)
                    rand_individuals = problem.generate_random_pop(pop_size - 1)
                    rand_fitness = np.array(list(map(problem.evaluate_fitness, rand_individuals)))
                    next_population = np.concatenate((best, rand_individuals))
                    next_fitness = np.concatenate((best_fit, rand_fitness))
                    nfev += len(rand_individuals)
                    tot_nfev += len(rand_individuals)
    
                    # Replace the population with a new population
                    population[:] = next_population
                    fitness[:] = next_fitness
    
                    # Re-compute the threshold and the decrement factor
                    thr, dec = self.initialize_cx_threshold(population, fitness)
    
            # Store the best solution ever found
            best_tracker.update(population, fitness)
    
            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res



import random
import numpy as np
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult
from typing import Union
from evovaq.problem import Problem


class DE(object):
    """
    Differential Evolution (DE) is a parallel direct search method designed for minimizing possibly nonlinear and
    non-differentiable continuous space functions [1]. Starting from an initial random population, the evolutionary
    cycle that creates a new population for the next generation consists of an appropriate perturbation of each individual.
    In detail, for each target individual, a mutant individual is generated by adding a weighted difference between two or four
    randomly chosen individuals to a third individual selected as best or randomly. At this point, a uniform crossover
    is applied between the target and the mutant to obtain the trial individual. Lastly, the trial individual is compared
    to the target individual to determine the best member of the next generation [2].

    References:
        [1] R. Storn and K. Price, “Differential evolution–a simple and efficient heuristic for global optimization over
        continuous spaces,” Journal of global optimization, vol. 11, no. 4, pp. 341–359, 1997.

        [2] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "A comparison of evolutionary algorithms for
        training variational quantum classifiers", Proceedings of 2023 IEEE Congress on Evolutionary Computation (CEC),
        pp. 1–8, 2023.

    Args:
        variant: The variants are denoted as <selection>/<n_diffs>/<crossover>, where <selection> is `best` or `rand`,
                 <n_diffs> is `1` or `2`, and <crossover> is `bin` or `exp`.
        differential_weight: Differential weight is a hyperparameter that controls the amplification of the individuals
                             difference used to create the mutant. If float, it must be in the range [0, 2].
                             If ``(min, max)`` tuple, dithering is employed. Dithering randomly changes the
                             differential weight in each generation so that it can help speed convergence significantly.
        CR: The probability that a parameter of the trial individual is replaced with the corresponding one of the
            mutant individual.
    """

    def __init__(self, variant: str = 'best/1/bin', differential_weight: Union[float, tuple] = 0.8, CR: float = 0.9):
        self.variant = variant
        self.differential_weight = differential_weight
        self.CR = CR

    def mut_de(self, population: np.ndarray, fitness: np.ndarray, target_idx: int, F: float) -> np.ndarray:
        """
        Create the mutant individual.

        Args:
            population: A population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                     shape.
            target_idx: Index of the target individual in the population.
            F: Differential weight.

        Returns:
            Mutant individual as array of real parameters with (`n_params`,) shape.
        """
        indices = np.arange(len(population))
        if self.variant.split('/')[0].lower() == 'rand':
            mutant_idx = np.random.choice(indices[indices != target_idx])
        elif self.variant.split('/')[0].lower() == 'best':
            mutant_idx = np.argsort(fitness)[0]
        else:
            raise ValueError("Please select a valid mutation strategy, i.e. 'best' or 'rand'")

        mutant = population[mutant_idx].copy()
        diff_ind_idx = np.random.choice(indices[~np.isin(indices, [target_idx, mutant_idx])],
                                        size=int(self.variant.split('/')[1]) * 2, replace=False)
        if self.variant.split('/')[1] == '1':
            mutated = mutant + F * (population[diff_ind_idx[0]].copy() - population[diff_ind_idx[1]].copy())
        elif self.variant.split('/')[1] == '2':
            mutated = mutant + F * (population[diff_ind_idx[0]].copy() + population[diff_ind_idx[1]].copy() -
                                    population[diff_ind_idx[2]].copy() - population[diff_ind_idx[3]].copy())
        else:
            raise ValueError("Please select a valid mutation strategy, i.e. '1' or '2'")
        return mutated

    def cx_de(self, target: np.ndarray, mutated: np.ndarray) -> np.ndarray:
        """
        Mate the target and mutant individual.

        Args:
            target: Target individual as array of real parameters with (`n_params`,) shape.
            mutated: Mutant individual as array of real parameters with (`n_params`,) shape.

        Returns:
            Trial individual as array of real parameters with (`n_params`,) shape.
        """
        variant_type = self.variant.split('/')[-1].lower()
        trial = None
        if variant_type == 'bin':
            R = np.random.choice(len(target))
            pr = np.random.random(size=len(target))
            trial = np.array([mutated[i] if pr[i] < self.CR or i == R else target[i] for i in range(len(target))])
        elif variant_type == 'exp':
            print("'exp' crossover is under construction")
        else:
            raise ValueError("Please select a valid crossover strategy, i.e. 'bin' or 'exp'")
        return trial

    def evolve_population(
            self, problem: Problem, population: np.ndarray, fitness: np.ndarray) -> tuple[np.ndarray, np.ndarray, int]:
        """
        Evolve the population by means of stochastic operators.

        Args:
            problem : :class:`~.Problem` to be solved.
            population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                    shape.

        Returns:
            The offspring and fitness values obtained after evolution, and number of fitness evaluations completed
            during the evolution.
        """

        # Check the dithering
        if isinstance(self.differential_weight, tuple):
            _min, _max = sorted(self.differential_weight)
            F = random.uniform(_min, _max)
        elif isinstance(self.differential_weight, float):
            F = self.differential_weight
        else:
            raise ValueError('The differential_weight must be a float in '
                             '(0, 2), or specified as a tuple(min, max)'
                             ' where min < max and min, max are in (0, 2).')

        # Clone the individuals to produce the offspring
        offspring = population.copy()
        fit_offspring = fitness.copy()

        for target_idx, child in enumerate(offspring):
            # Mutation step
            mutated = self.mut_de(population, fitness, target_idx, F)

            # Crossover step
            trial = self.cx_de(child, mutated)

            # Check the bounds of the parameters
            trial[:] = problem.check_bounds(trial)

            # Selection step
            fit_trial = problem.evaluate_fitness(trial)
            if fit_trial < fit_offspring[target_idx]:
                child[:] = trial
                fit_offspring[target_idx] = fit_trial
        nfev = len(offspring)
        return offspring, fit_offspring, nfev

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            offspring, fit_offspring, nfev = self.evolve_population(problem, population, fitness)
            tot_nfev += nfev

            population[:] = offspring
            fitness[:] = fit_offspring

            # Store the best solution ever found
            best_tracker.update(population, fitness)

            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res



import numpy as np
from typing import Union, Sequence


# ---------- SELECTION OPERATORS ------------
def sel_best(population: np.ndarray, fitness: np.ndarray, k: int) -> tuple[np.ndarray, np.ndarray]:
    """
    Select the first `k` best individuals (with the smallest objective values) in the population.

    Args:
        population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                    shape.
        fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                shape.
        k: Number of individuals to be selected.

    Returns:
        Subsets of `k` best individuals and fitness values.
    """
    sorted_idx = np.argsort(fitness)
    return population[sorted_idx[:k]], fitness[sorted_idx[:k]]


def sel_permutation(population: np.ndarray, fitness: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Select the mating pool by permutation of the population indices.

    Args:
        population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                    shape.
        fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                shape.

    Returns:
        Mating pool with individuals and fitness values with new indices.
    """
    permutation_idx = np.random.permutation(len(population))
    return population[permutation_idx], fitness[permutation_idx]


def sel_tournament(population: np.ndarray, fitness: np.ndarray, k: int, tournsize: int = 3) -> tuple[np.ndarray, np.ndarray]:
    """
    Select the best individual among `tournsize` randomly chosen individuals, `k` times.

    Args:
        population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                    shape.
        fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                shape.
        k: Number of individuals to be selected.
        tournsize:  Number of random individuals participating in each tournament.

    Returns:
        Subsets of `k` individuals and fitness values.
    """
    chosen = []
    for _ in range(k):
        aspirants = np.random.choice(len(population), size=tournsize)
        fit_aspirants = fitness[aspirants]
        winner = aspirants[np.argmin(fit_aspirants)]
        chosen.append(winner)
    return population[chosen], fitness[chosen]


def sel_random(population: np.ndarray, fitness: np.ndarray, k: int, replace: bool = True) -> tuple[np.ndarray, np.ndarray]:
    """
    Select `k` individuals randomly.

    Args:
        population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                    shape.
        fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                shape.
        k: Number of individuals to be selected.
        replace: Whether the sample is with or without replacement. Default is True, meaning that an element can be
                 selected multiple times.

    Returns:
        Subsets of `k` individuals and fitness values randomly chosen.
    """
    random_idx = np.random.choice(len(population), size=k, replace=replace)
    return population[random_idx], fitness[random_idx]


# ---------- CROSSOVER OPERATORS ------------
def cx_blx_alpha(par1: np.ndarray, par2: np.ndarray, alpha: float = 0.5) -> tuple[np.ndarray, np.ndarray]:
    """
    BLX-alpha crossover.

    Args:
        par1: The first parent as array of real parameters with (`n_params`,) shape.
        par2: The second parent as array of real parameters with (`n_params`,) shape.
        alpha: Positive real hyperparameter.

    Returns:
        The two resulting children.
    """
    d = np.abs(par1 - par2)
    stacked = np.stack((par1, par2), axis=1)
    lower_bounds = np.min(stacked, axis=1) - alpha * d
    upper_bounds = np.max(stacked, axis=1) + alpha * d
    child1 = np.random.uniform(lower_bounds, upper_bounds, par1.shape)
    child2 = np.random.uniform(lower_bounds, upper_bounds, par2.shape)
    return child1, child2


def cx_one_point(par1: np.ndarray, par2: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    One-point crossover.

    Args:
        par1: The first parent as array of real parameters with (`n_params`,) shape.
        par2: The second parent as array of real parameters with (`n_params`,) shape.

    Returns:
        The two resulting children.
    """
    point = np.random.randint(1, par1.size - 1)
    child1 = np.concatenate((par1[:point], par2[point:]))
    child2 = np.concatenate((par2[:point], par1[point:]))
    return child1, child2


def cx_two_point(par1: np.ndarray, par2: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Two-point crossover.

    Args:
        par1: The first parent as array of real parameters with (`n_params`,) shape.
        par2: The second parent as array of real parameters with (`n_params`,) shape.

    Returns:
        The two resulting children.
    """
    point1, point2 = sorted(np.random.choice(par1.size, 2, replace=False))
    child1 = np.concatenate((par1[:point1], par2[point1:point2], par1[point2:]))
    child2 = np.concatenate((par2[:point1], par1[point1:point2], par2[point2:]))
    return child1, child2


def cx_uniform(par1: np.ndarray, par2: np.ndarray, cx_indpb: float = 0.5) -> tuple[np.ndarray, np.ndarray]:
    """
    Uniform crossover.

    Args:
        par1: The first parent as array of real parameters with (`n_params`,) shape.
        par2: The second parent as array of real parameters with (`n_params`,) shape.
        cx_indpb: Independent probability for each parameter to be exchanged.

    Returns:
        The two resulting children.
    """
    mask = np.random.choice([True, False], size=par1.size, p=[cx_indpb, 1 - cx_indpb])
    child1 = np.where(mask, par1, par2)
    child2 = np.where(mask, par2, par1)
    return child1, child2


# ---------- MUTATION OPERATORS ------------
def mut_gaussian(individual: np.ndarray, mu: Union[float, Sequence] = 0, sigma: Union[float, Sequence] = 1,
                 mut_indpb: float = 0.1) -> np.ndarray:
    """
    Gaussian mutation.

    Args:
        individual: Individual to be mutated as array of real parameters with (`n_params`,) shape.
        mu: Mean or sequence of means for the gaussian addition mutation.
        sigma: Standard deviation or sequence of standard deviations for the gaussian addition mutation.
        mut_indpb: Independent probability for each parameter to be mutated.

    Returns:
        Resulting mutated individual.
    """
    mask = np.random.choice([True, False], size=individual.size, p=[mut_indpb, 1 - mut_indpb])
    mutated_individual = individual + mask * np.random.normal(mu, sigma, size=individual.shape)
    return mutated_individual


def mut_flip_bit(individual: np.ndarray, mut_indpb: float = 0.1) -> np.ndarray:
    """
    Flip-bit mutation.

    Args:
        individual: Individual to be mutated as array of real parameters with (`n_params`,) shape.
        mut_indpb: Independent probability for each parameter to be mutated.

    Returns:
        Resulting mutated individual.
    """
    mutated_individual = individual.copy()
    mask = np.random.choice([True, False], size=individual.size, p=[mut_indpb, 1 - mut_indpb])
    mutated_individual[mask] = 1 - mutated_individual[mask]
    return mutated_individual


SELECTION_ARGS = {
    sel_tournament: ['tournsize']
}

CROSSOVER_ARGS = {
    cx_blx_alpha: ['alpha'],
    cx_uniform: ['cx_indpb']
}

MUTATION_ARGS = {
    mut_gaussian: ['mu', 'sigma', 'mut_indpb'],
    mut_flip_bit: ['mut_indpb']
}


from setuptools import find_packages, setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='evovaq',
    packages=['evovaq', 'evovaq.tools'],
    version='1.0.21',
    description='EVOlutionary algorithms toolbox for VAriational Quantum circuits',
    author='Angela Chiatto',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='angela.chiatto@unina.it',
    license='MIT',
    url='https://github.com/Quasar-UniNA/EVOVAQ',
    keywords=['Quantum Computing', 'Evolutionary Algorithms', 'Variational Quantum Circuits'],
    install_requires=[
        'numpy==1.23.5',
        'tabulate==0.8.10',
        'tqdm==4.64.1',
        'matplotlib==3.5.1',
        'pandas==1.4.2',
        'openpyxl==3.0.9'
      ],
    classifiers=[
        'Development Status :: 3 - Alpha',      
        'Intended Audience :: Developers',      
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3'
      ],
)



import numpy as np
from typing import Union


def euclidean(a: Union[float, np.ndarray], b: Union[float, np.ndarray]) -> float:
    """
    Euclidean distance between two points.

    Args:
        a: First point.
        b: Second point.

    Returns:
        Euclidean distance.
    """
    return np.linalg.norm(a - b)



import random
import numpy as np
from functools import partial
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult
from evovaq.tools.operators import sel_best
from evovaq.tools.operators import SELECTION_ARGS, CROSSOVER_ARGS, MUTATION_ARGS
from typing import Callable, Union
from evovaq.problem import Problem


class GA(object):

    """
    Genetic Algorithm (GA) is the simplest evolutionary algorithm inspired by Darwinian evolution principles: the evolution
    of a population of possible solutions to a given problem is linked to the concepts of randomness and survival of the
    fittest [1]. In detail, during an evolutionary cycle, the processes of selection, crossover and mutation, known as
    stochastic genetic operators, take place in order to produce the next generation of solutions. The evolution process
    stops when a maximum number of generations or fitness evaluations is reached. Typically, during the evolution
    process, the best individual of the current generation can be inserted into the next one in order to prevent its
    possible disappearance [2].

    References:
        [1] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "Training variational quantum circuits through
        genetic algorithms", Proceedings of 2022 IEEE Congress on Evolutionary Computation (CEC), pp. 1–8, 2022.

        [2] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "Genetic algorithms as classical optimizer for the
        quantum approximate optimization algorithm", Applied Soft Computing, pp. 110296, Elsevier, 2023.

    Args:
        selection: Selection operator used to select individuals to create the next generation. The selection function
                   is defined as ``sel_function(pop, fitness, *args) -> (sel_pop, sel_fit)``, where ``pop`` and ``fitness``
                   are two arrays of real parameters with (`pop_size`,`n_params`) and (`n_params`,) shape respectively,
                   ``args`` is a tuple of other fixed parameters needed to specify the function, and the output is the
                   resulting subset of individuals and fitness values.
        crossover: Crossover operator used to mate two individuals. The crossover function is defined as
                   ``cx_function(par1, par2, *args) -> (child1, child2)``, where ``par1`` and ``par2`` are two arrays of
                   real parameters with (`n_params`,) shape, ``args`` is a tuple of other fixed parameters needed to
                   specify the function, and the output is the resulting children with the same shape as the parents.
        mutation: Mutation operator used to mutate individuals. The mutation function is defined as
                   ``mut_function(ind, *args) -> mutated_ind``, where ``ind`` is an array of real parameters with
                   (`n_params`,) shape, ``args`` is a tuple of other fixed parameters needed to specify the function,
                   and the output is the mutated individual.
        elitism: If True, the best solution of current population is transferred directly into the next generation.
        cxpb: The probability of mating two individuals.
        mutpb: The probability of mutating an individual.
        kwargs: Additional keyword arguments used to set hyperparameter values of genetic operators.
    """

    def __init__(
            self, selection: Callable, crossover: Callable, mutation: Callable, elitism: bool = True,
            cxpb: float = 0.8, mutpb: float = 1.0, **kwargs):
        self.elitism = elitism
        self.cxpb = cxpb
        self.mutpb = mutpb
        self.kwargs = kwargs

        if selection in SELECTION_ARGS:
            args = {arg: self.kwargs.get(arg) for arg in SELECTION_ARGS[selection] if arg in self.kwargs.keys()}
            self.selection = partial(selection, **args)
        else:
            self.selection = selection

        if crossover in CROSSOVER_ARGS:
            args = {arg: self.kwargs.get(arg) for arg in CROSSOVER_ARGS[crossover] if arg in self.kwargs.keys()}
            self.crossover = partial(crossover, **args)
        else:
            self.crossover = crossover

        if mutation in MUTATION_ARGS:
            args = {arg: self.kwargs.get(arg) for arg in MUTATION_ARGS[mutation] if arg in self.kwargs.keys()}
            self.mutation = partial(mutation, **args)
        else:
            self.mutation = mutation

    def evolve_population(
            self, problem: Problem, population: np.ndarray, fitness: np.ndarray) -> tuple[np.ndarray, np.ndarray, int]:

        """
        Evolve the population by means of genetic operators.

        Args:
            problem : :class:`~.Problem` to be solved.
            population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                    shape.

        Returns:
            The offspring and fitness values obtained after evolution, and number of fitness evaluations completed
            during the evolution.
        """

        # Select the next generation individuals
        parents, fit_parents = self.selection(population=population, fitness=fitness, k=len(population))

        # Clone the individuals to produce the offspring
        offspring = parents.copy()
        fit_offspring = fit_parents.copy()
        recompute_fitness = np.zeros(len(fitness), dtype=bool)

        # Crossover step
        for i, (child1, child2) in enumerate(zip(offspring[::2], offspring[1::2])):
            if random.random() < self.cxpb:
                child1[:], child2[:] = self.crossover(child1, child2)
                recompute_fitness[i] = recompute_fitness[i + 1] = True

        # Mutation step
        for i, mutant in enumerate(offspring):
            if random.random() < self.mutpb:
                mutant_clone = mutant.copy()
                mutant[:] = self.mutation(mutant)
                if not np.array_equal(mutant_clone, mutant):
                    recompute_fitness[i] = True

        # Check the bounds of the parameters
        for child in offspring:
            child[:] = problem.check_bounds(child)

        # Evaluate the individuals with an invalid fitness
        fit_offspring[recompute_fitness] = np.array(list(map(problem.evaluate_fitness, offspring[recompute_fitness])))
        nfev = len(fit_offspring[recompute_fitness])
        return offspring, fit_offspring, nfev

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            best, best_fit = sel_best(population, fitness, 1)

            # Evolve the population based on the genetic operators chosen
            offspring, fit_offspring, nfev = self.evolve_population(problem, population, fitness)
            tot_nfev += nfev

            if self.elitism:
                joint_pop = np.concatenate((offspring, best))
                joint_fitness = np.concatenate((fit_offspring, best_fit))

                # Sort from the best (min fitness) to worst (max fitness) individual
                sorted_idx = np.argsort(joint_fitness)

                # Replacement
                population[:] = joint_pop[sorted_idx[:pop_size]]
                fitness[:] = joint_fitness[sorted_idx[:pop_size]]
            else:
                population[:] = offspring
                fitness[:] = fit_offspring

            # Store the best solution ever found
            best_tracker.update(population, fitness)

            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res



import numpy as np
import random
from evovaq.tools.support import compute_statistics, print_info, Logbook, set_progress_bar, BestIndividualTracker, \
    FinalResult, Particle
from evovaq.tools.operators import sel_best
from typing import Union
from evovaq.problem import Problem


class PSO(object):
    """
    Particle Swarm Optimization (PSO) is an optimization method based on a swarm of candidate solutions, named particles,
    moving in the search space according to appropriate position and velocity equation [1]. Starting from a swarm of
    particles with random positions and velocities, each particle’s velocity is updated by combining its own best
    position (pbest) and the global best position (gbest) ever found in the search space with some random perturbations
    influenced by two hyperparameters, denoted as `phi1` and `phi2`. At this point, each particle’s position is updated
    by adding its resulting velocity to the current position. The final goal is to move the whole swarm close to an
    optimal position in the search space.
    In `evovaq`, PSO with inertia weight (`inertia_weight`) described in [2] is implemented.

    References:
        [1] Giovanni Acampora, Angela Chiatto, and Autilia Vitiello, "A comparison of evolutionary algorithms for
        training variational quantum classifiers", Proceedings of 2023 IEEE Congress on Evolutionary Computation (CEC),
        pp. 1–8, 2023.

        [2] Poli R., Kennedy J., & Blackwell T., "Particle swarm optimization: An overview", Swarm intelligence, vol. 1,
        pp. 33-57, 2007.

    Args:
        vmin: Lower value(s) of the velocity. If None, no limits are considered.
        vmax: Upper value(s) of the velocity. If None, no limits are considered.
        inertia_weight: Inertia weight.
        phi1: Acceleration coefficient determining the magnitude of the random force in the direction of personal
              best solution.
        phi2: Acceleration coefficient determining the magnitude of the random force in the direction of global best
              solution.
    """
    def __init__(self, vmin: Union[float, np.ndarray, None] = None, vmax: Union[float, np.ndarray, None] = None,
                 inertia_weight: float = 0.7298, phi1: float = 1.49618, phi2: float = 1.49618):
        self.vmin = vmin
        self.vmax = vmax
        self.inertia_weight = inertia_weight
        self.phi1 = phi1
        self.phi2 = phi2

    @staticmethod
    def initialize_swarm(population: np.ndarray, fitness: np.ndarray) -> list[Particle]:
        """
        Initialize the swarm.

        Args:
            population: A population of individuals as array of real parameters with (`pop_size`, `n_params`)
                        shape.
            fitness: A set of fitness values associated to the population as array of real values with (`pop_size`,)
                     shape.

        Returns:
            List of :class:`~.Particle` elements.
        """
        swarm = []
        for position, fit in zip(population, fitness):
            velocity = np.random.uniform(-1, 1, len(position))
            part = Particle(position, velocity, fit)
            swarm.append(part)
        return swarm

    def optimize(
            self, problem: Problem, pop_size: int, initial_pop: Union[np.ndarray, None] = None,
            max_nfev: Union[int, None] = None, max_gen: int = 1000, n_run: int = 1, seed: Union[int, float, str, None] = None,
            verbose: bool = True) -> FinalResult:
        """
        Optimize the parameters of the problem to be solved.

        Args:
            problem: :class:`~.Problem` to be solved.
            pop_size: Population size.
            initial_pop: Initial population of possible solutions as array of real parameters with (`pop_size`, `n_params`)
                         shape. If None, the initial population is randomly generated from `param_bounds`.
            max_nfev: Maximum number of fitness evaluations used as stopping criterion. If None, the maximum number of
                      generations `max_gen` is considered as stopping criterion.
            max_gen: Maximum number of generations used as stopping criterion. If `max_nfev` is not None, this is
                     considered as stopping criterion.
            n_run: Independent execution number of the algorithm.
            seed: Initialize the random number generator. If None, the current time is used.
            verbose: If True, the statistics of fitness values is printed during the evolution.

        Returns:
            A :class:`~.FinalResult` containing the optimization result.
        """

        # Set the seed
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        # Create and initialize the population
        if initial_pop is None:
            population = problem.generate_random_pop(pop_size)
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)
        else:
            population = initial_pop[:].copy()
            fitness = np.array(list(map(problem.evaluate_fitness, population)))
            nfev = len(population)
            tot_nfev = len(population)

        # Store the best solution ever found
        best_tracker = BestIndividualTracker()
        best_tracker.update(population, fitness)

        # Set the progress bar considering the stopping criterion
        pbar = set_progress_bar(max_gen, max_nfev)
        if max_nfev is not None:
            pbar.update(nfev)

        # Compute the statistics of the fitness values
        stats = compute_statistics(fitness)

        # Set the logbook
        lg = Logbook()
        lg.record(gen=0, nfev=nfev, **stats)

        # Print the evolution info
        if verbose:
            print_info(n_run=n_run, gen=0, nfev=nfev, **stats, header=True)

        swarm = self.initialize_swarm(population, fitness)

        # Begin the generational process
        for gen in range(1, max_gen + 1):

            # Check the stopping criterion
            if max_nfev is not None and tot_nfev >= max_nfev:
                pbar.close()
                break

            gbest, fit_gbest = sel_best(population, fitness, 1)
            offspring = population.copy()
            fit_offspring = fitness.copy()

            for idx, part in enumerate(swarm):
                part.update_velocity(gbest[0], self.inertia_weight, self.phi1, self.phi2, self.vmin, self.vmax)
                part.update_position(problem.param_bounds)
                new_fitness = problem.evaluate_fitness(part.position)
                part.update_fitness(new_fitness)
                part.update_pbest()
                offspring[idx] = part.position
                fit_offspring[idx] = part.fitness

            nfev = len(offspring)
            tot_nfev += len(offspring)

            population[:] = offspring
            fitness[:] = fit_offspring

            # Store the best solution ever found
            best_tracker.update(population, fitness)

            # Compute the statistics of the fitness values
            stats = compute_statistics(fitness)

            # Record info in the logbook
            lg.record(gen=gen, nfev=nfev, **stats)

            # Print the evolution info
            if verbose:
                print_info(n_run=n_run, gen=gen, nfev=nfev, **stats)

            # Update the progress bar
            if max_nfev is not None:
                pbar.update(nfev)
            else:
                pbar.update()

        res = FinalResult(x=best_tracker.get_best(), fun=best_tracker.get_best_fit(), nfev=tot_nfev, gen=gen,
                          log=lg.get_log())
        return res
